@@ -525,13 +525,8 @@ wl_text_input_manager_create_text_input(struct wl_text_input_manager *wl_text_in
 {
 	struct wl_proxy *id;
 
-	id = wl_proxy_create((struct wl_proxy *) wl_text_input_manager,
-			     &wl_text_input_interface);
-	if (!id)
-		return NULL;
-
-	wl_proxy_marshal((struct wl_proxy *) wl_text_input_manager,
-			 WL_TEXT_INPUT_MANAGER_CREATE_TEXT_INPUT, id);
+	id = wl_proxy_marshal_constructor((struct wl_proxy *) wl_text_input_manager,
+			 WL_TEXT_INPUT_MANAGER_CREATE_TEXT_INPUT, &wl_text_input_interface, NULL);
 
 	return (struct wl_text_input *) id;
 }
