@@ -27,6 +27,14 @@
 #include <errno.h>
 #include <linux/input.h>
 
+#if HAVE_FREERDP_VERSION_H
+#include <freerdp/version.h>
+#else
+/* assume it's a early 1.1 version */
+#define FREERDP_VERSION_MAJOR 1
+#define FREERDP_VERSION_MINOR 1
+#endif
+
 #include <freerdp/freerdp.h>
 #include <freerdp/listener.h>
 #include <freerdp/update.h>
@@ -302,7 +310,7 @@ rdp_output_start_repaint_loop(struct weston_output *output)
 	weston_output_finish_frame(output, msec);
 }
 
-static void
+static int
 rdp_output_repaint(struct weston_output *output_base, pixman_region32_t *damage)
 {
 	struct rdp_output *output = container_of(output_base, struct rdp_output, base);
@@ -324,6 +332,7 @@ rdp_output_repaint(struct weston_output *output_base, pixman_region32_t *damage)
 				 &ec->primary_plane.damage, damage);
 
 	wl_event_source_timer_update(output->finish_frame_timer, 16);
+	return 0;
 }
 
 static void
@@ -485,8 +494,6 @@ rdp_compositor_create_output(struct rdp_compositor *c, int width, int height,
 	if (pixman_renderer_output_create(&output->base) < 0)
 		goto out_shadow_surface;
 
-	weston_output_move(&output->base, 0, 0);
-
 	loop = wl_display_get_event_loop(c->base.wl_display);
 	output->finish_frame_timer = wl_event_loop_add_timer(loop, finish_frame_handler, output);
 
@@ -522,7 +529,6 @@ rdp_restore(struct weston_compositor *ec)
 static void
 rdp_destroy(struct weston_compositor *ec)
 {
-	ec->renderer->destroy(ec);
 	weston_compositor_shutdown(ec);
 
 	free(ec);
@@ -573,7 +579,11 @@ rdp_peer_context_new(freerdp_peer* client, RdpPeerContext* context)
 	context->item.peer = client;
 	context->item.flags = RDP_PEER_OUTPUT_ENABLED;
 
+#if FREERDP_VERSION_MAJOR == 1 && FREERDP_VERSION_MINOR == 1
 	context->rfx_context = rfx_context_new();
+#else
+	context->rfx_context = rfx_context_new(TRUE);
+#endif
 	context->rfx_context->mode = RLGR3;
 	context->rfx_context->width = client->settings->DesktopWidth;
 	context->rfx_context->height = client->settings->DesktopHeight;
@@ -598,8 +608,11 @@ rdp_peer_context_free(freerdp_peer* client, RdpPeerContext* context)
 			wl_event_source_remove(context->events[i]);
 	}
 
-	if(context->item.flags & RDP_PEER_ACTIVATED)
+	if(context->item.flags & RDP_PEER_ACTIVATED) {
+		weston_seat_release_keyboard(&context->item.seat);
+		weston_seat_release_pointer(&context->item.seat);
 		weston_seat_release(&context->item.seat);
+	}
 	Stream_Free(context->encode_stream, TRUE);
 	nsc_context_free(context->nsc_context);
 	rfx_context_free(context->rfx_context);
