@@ -41,12 +41,10 @@ enum evdev_event_type {
 	EVDEV_RELATIVE_MOTION,
 };
 
-enum evdev_device_capability {
-	EVDEV_KEYBOARD = (1 << 0),
-	EVDEV_BUTTON = (1 << 1),
-	EVDEV_MOTION_ABS = (1 << 2),
-	EVDEV_MOTION_REL = (1 << 3),
-	EVDEV_TOUCH = (1 << 4),
+enum evdev_device_seat_capability {
+	EVDEV_SEAT_POINTER = (1 << 0),
+	EVDEV_SEAT_KEYBOARD = (1 << 1),
+	EVDEV_SEAT_TOUCH = (1 << 2)
 };
 
 struct evdev_device {
@@ -55,11 +53,14 @@ struct evdev_device {
 	struct wl_event_source *source;
 	struct weston_output *output;
 	struct evdev_dispatch *dispatch;
+	struct wl_listener output_destroy_listener;
 	char *devnode;
 	char *devname;
+	char *output_name;
 	int fd;
 	struct {
 		int min_x, max_x, min_y, max_y;
+		uint32_t seat_slot;
 		int32_t x, y;
 
 		int apply_calibration;
@@ -70,6 +71,7 @@ struct evdev_device {
 		int slot;
 		struct {
 			int32_t x, y;
+			uint32_t seat_slot;
 		} slots[MAX_SLOTS];
 	} mt;
 	struct mtdev *mtdev;
@@ -79,7 +81,7 @@ struct evdev_device {
 	} rel;
 
 	enum evdev_event_type pending_event;
-	enum evdev_device_capability caps;
+	enum evdev_device_seat_capability seat_caps;
 
 	int is_mt;
 };
@@ -122,6 +124,9 @@ evdev_led_update(struct evdev_device *device, enum weston_led leds);
 struct evdev_device *
 evdev_device_create(struct weston_seat *seat, const char *path, int device_fd);
 
+void
+evdev_device_set_output(struct evdev_device *device,
+			struct weston_output *output);
 void
 evdev_device_destroy(struct evdev_device *device);
 
